package info.textgrid.services.aggregator;

import info.textgrid.services.aggregator.util.StylesheetManager;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.util.List;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.StreamingOutput;
import javax.ws.rs.core.UriBuilder;
import javax.ws.rs.core.UriInfo;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.s9api.QName;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.Serializer;
import net.sf.saxon.s9api.XdmAtomicValue;
import net.sf.saxon.s9api.XsltTransformer;

import com.google.common.base.Optional;
import com.google.common.net.HostAndPort;

public class Help implements StreamingOutput {

	private final StylesheetManager stylesheetManager;
	private UriInfo uriInfo;
	private HttpHeaders headers;

	public Help(final StylesheetManager stylesheetManager) {
		this.stylesheetManager = stylesheetManager;
	}

	@Override
	public void write(final OutputStream output) throws IOException,
			WebApplicationException {
		try {
			final XsltTransformer transformer = stylesheetManager
					.getStylesheet(
							URI.create("/stylesheets/wadl_documentation-2009-02.xsl"),
							Optional.<String> absent(), false, false).load();
			final Serializer serializer = stylesheetManager.xsltProcessor
					.newSerializer(output);
			transformer.setDestination(serializer);
			transformer.setSource(new StreamSource(getWadlURL().openStream(),
					getWadlURL().toString()));
			transformer.setParameter(new QName("apptitle"), new XdmAtomicValue(
					"TextGrid Aggregator Service"));
			transformer.transform();
		} catch (final SaxonApiException e) {
			throw new WebApplicationException(e);
		} catch (final FileNotFoundException e) {
			throw new WebApplicationException(GenericExceptionMapper
					.toResponse(
							Status.NOT_FOUND,
							"There still was a not found error: "
									+ e.getMessage(),
							"Path:      " + uriInfo.getPath() + "\nBase URI:  "
									+ uriInfo.getBaseUri().toString()
									+ "\nAbsolute Path: "
									+ uriInfo.getAbsolutePath()
									+ "\nRequest URI: "
									+ uriInfo.getRequestUri().toString())
					.build());
		}

	}

	public Help uriInfo(UriInfo uriInfo) {
		this.uriInfo = uriInfo;
		return this;
	}

	public Help headers(HttpHeaders headers) {
		this.headers = headers;
		return this;
	}

	public URL getWadlURL() {
		if (headers != null) {
			final List<String> originalUri = headers
					.getRequestHeader("X-Original-Request-URI");
			if (originalUri != null && originalUri.size() > 0) {
				final URL wadlURL = getWadlURL(originalUri.get(0));
				if (wadlURL != null)
					return wadlURL;
			}
		}
		if (uriInfo != null) {
			URL wadlURL = getWadlURL(uriInfo.getAbsolutePath().toString());
			if (wadlURL != null)
				return wadlURL;
			else if (headers != null) {
				// Somehow uriInfo fails to recognize the port from the Host header, but
				// only on the production machines. This is a workaround.
				final HostAndPort host = HostAndPort.fromString(
						headers.getRequestHeader("Host").get(0))
						.withDefaultPort(80);
				wadlURL = getWadlURL(uriInfo.getAbsolutePathBuilder().host(host.getHostText())
						.port(host.getPort())
						.scheme(host.getPort() == 443 ? "https" : "http")
						.build().toString());
				if (wadlURL != null)
					return wadlURL;
			}
		}
		throw new WebApplicationException(GenericExceptionMapper.toResponse(
				Status.BAD_REQUEST,
				"Unable to determine a valid path to the WADL.",
				"Path:      " + uriInfo.getPath() + "\nBase URI:  "
						+ uriInfo.getBaseUri().toString() + "\nAbsolute Path: "
						+ uriInfo.getAbsolutePath() + "\nRequest URI: "
						+ uriInfo.getRequestUri().toString()).build());
	}

	private URL getWadlURL(String string) {
		final URI resolved = UriBuilder.fromUri(string).replaceQuery("_wadl")
				.build();
		try {
			final URL url = resolved.toURL();
			final URLConnection connection = url.openConnection();
			connection.connect();
			if (((HttpURLConnection) connection).getResponseCode() == 200)
				return url;
			else
				return null;
		} catch (final MalformedURLException e) {
			throw new WebApplicationException(e);
		} catch (final IOException e) {
			throw new WebApplicationException(e);
		}
	}
}